
/*
    a tap to intercept a shutdown
    - for 1209 only

    - when a user presses the remote OFF button, this tap will
      intercept it, and generate an RKEY_Power event to all taps

    - an interested tap may shut down nicely, save state etc
    - a non-interested tap should ignore RKEY_Power (return RKEY_Power in param1)

    - this tap will then generate a 2nd RKEY_Power event
    - following this, the box will shutdown

    -how does it work?
    on pressing the remote 'OFF' button, an IRQ is generated
    the IRQ service routine generates the shutdown task, task_id = 0xef00
    the OS will then schedule this task, entry point: 0x80108c0c
    the shutdown task
        - disables something irq (watchdog?)
        - displays 'OFF' on the front panel
        - checks to see if a recording is taking place; if so pops up an 'are you sure?' dialog
        - stops recording
        - stops the playback
        - jumps into a dead loop (0x80108df4)
        - shutsdown (presumably due to the watchdog timing out)

    this tap hooks the shutdown event


    I think that the initial IRQ may enable a watchdog, as once the 'OFF' key is pressed,
    then the box will always shutdown after a few seconds unless the 'disable something irq' routine is called.

    However, once the 'disable something irq' is called, further presses of the 'OFF' button get ignored and
    do not cause a shutdown task.

    So, there is a bit of jiggery pokery needed to get this to work.

    1. press 'off'
    2. get 1st shutdown event hook, disable irq, generate RKEY_Power event to all taps, return
    3. on 1st RKEY_Power event, generate another RKEY_Power event
    4. get 2nd shutdown event hook, return  (sometimes get this one before the RKEY_Power event activates)
    5. on 2nd RKEY_Power, renable irq, generate shutdown task
    6. get 3rd shutdown event hook, run original task code; box will shutdown


    a user tap simply needs to monitor for the RKEY_Power event
    (and not expect to get any idle events after receiving it)
    (and ignore it if it receives it a second time)
    or just ignore it

    a user tap will also receive the hook event (event=0xf100). It must ignore it and MUST return it with param1 unchanged.

    ... note the hook functions can be used for lots of interesting things
    e.g.
        intercepting _any_ firmware functions
        grabbing the raw TS data
        etc, etc


    - this tap also demonstrates shutting down the box
    - call the function
        exTAP_Shutdown();           to shutdown the box


    - this tap also demonstrates loading another tap
    - call the function
        exTAP_TAP_Load("filename.tap");     to load a tap


    - this tap also includes 'mheg_off'
    - call the function
        exTAP_MHEG_Disable(1);      to disable MHEG
        exTAP_MHEG_Disable(0);      to enable MHEG


    to use these functions, ensure the .h file is included
        #include "exTap.h"
    and the library 'exTap.a' is linked in, similarly to 'libtap.a'
*/
#include <tap.h>
#include "exTap.h"


//Every TAP application needs unique ID and TAP allocations that have same ID cant execute at same time.
//ID consists of DWORD( 4byte ) and the MSB bit should be set by 1.
//The following is how to use in source codes.
#define ID_USERTAP 0xbdb00401
TAP_ID( ID_USERTAP );

//Every TAP application has to include Program name, Author name, and Program description as following
//manners in source code.
TAP_PROGRAM_NAME("shutdown");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("A TAP to intercept a shutdown");
TAP_ETCINFO(__DATE__);

//----------------------------------------------------------------------------
typedef struct {
    int  id;
    char str[0x20];
}type_item;

typedef struct {
    int  id;
    int  visible;
    TYPE_Window win;
    int  rgn;
    char *fixme_titlep;
    type_item item[0x20];
}type_menu;

char menu_title[0x20] = "Shutdown";

type_menu menu = {
    0,
    0,
    {0},
    0,
    menu_title,
    {
        {8, "xxx",                      },
        {0, "Enable Shutdown Event",    },
        {1, "Disable Shutdown Event",   },
        {2, "Shutdown",                 },
        {3, "Issue RKEY_Power",         },
        {4, "Load a TAP",               },
        {5, "Disable MHEG",             },
        {6, "Enable MHEG",              },
        {7, "Exit Tap", },
    },
};

dword message_showing = 0;
dword message_rgn = 0;

dword ShowMessage(char* lpMessage, char* lpMessage1, dword dwDelay, int wait);
void ShowMenu(type_menu *menu);
void DeleteMenu(type_menu *menu);

//----------------------------------------------------------------------------
int TAP_Main(void);
dword TAP_EventHandler(word event, dword param1, dword param2);
int menu_handler(type_menu *menu, dword param1);
int key_handler(dword param1);
int init(void);
int close(void);

//----------------------------------------------------------------------------
extern type_menu load_tap_menu;
int load_tap_menu_fill();
int load_tap_menu_handler(type_menu *menu, dword param1);

//----------------------------------------------------------------------------
#define FW_SHUTDOWN_TASK_1209          0x80108c0c
#define FW_SHUTDOWN_IRQ_DISABLE_1209   0x80004da4
#define FW_SHUTDOWN_IRQ_ENABLE_1209    0x80004dd8

int hook_id[16] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,};

int in_shutdown = 0;
int apply_hooks(void);
int remove_hooks(void);
dword shutdown_hook(word event, dword param1, dword param2);

//----------------------------------------------------------------------------

int TAP_Main(void)
{
    int ret;

    TAP_Print("\nopening %s",__tap_program_name__);

    //determine support for different firmwares
    ret = exTAP_Init();
    if (ret) {
        return 0;
    }

    init();
    return 1;                 //0=application mode, 1=tsr mode
}

//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    int ret;
    dword state, subState;


    switch (event) {
        case EVT_IDLE:
//            TAP_Print("\nEVT_IDLE, event=%x, param1=%x, param2=%x", event, param1, param2);

            //remove a message when it times out
            if (message_showing) {
                if (TAP_GetTick() > message_showing) {
                	TAP_Osd_Delete(message_rgn);    // release rgn-handle
                    message_showing = 0;
                    message_rgn = 0;
                }
            }
            break;
        case EVT_KEY:
            TAP_GetState(&state, &subState);
            if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                if (menu.visible) { // handling keys when options window is shown
                    param1 = menu_handler(&menu, param1);
                }
                else
                if (load_tap_menu.visible) {
                    param1 = load_tap_menu_handler(&load_tap_menu, param1);
                }
                else {
                    param1 = key_handler(param1);
                }
            }
            break;
        case EVT_HOOK:
//            TAP_Print("\nEVT_HOOK, event=%x, param1=%x, param2=%x", event, param1, param2);
            if (param1 == hook_id[0]) {
                param1 = shutdown_hook(event, param1, param2);
            }
            break;
        default:
//TAP_Print("\nEVT_???, event=%x, param1=%x, param2=%x", event, param1, param2);
            break;
    }

    return param1;
}

int menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int ret;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:
				    ret = apply_hooks();
                    if (ret == 0) strcpy(menu_title, "Shutdown is Enabled");
				    return 0;
				case 1:
				    ret = remove_hooks();
                    if (ret == 0) strcpy(menu_title, "Shutdown is Disabled");
				    return 0;
				case 2:
                    TAP_Print("\nexTAP_Shutdown()");
                    exTAP_Shutdown();
				    return 0;
				case 3:
                    TAP_Print("\nTAP_GenerateEvent(EVT_KEY, RKEY_Power, 0)");
                    TAP_GenerateEvent(EVT_KEY, RKEY_Power, 0);
				    return 0;
			    case 4:
			        //fill in the menu
			        load_tap_menu_fill();
                    ShowMenu(&load_tap_menu);
				    return 0;
				case 5:
                    exTAP_MHEG_Disable(1);      //0=mheg on, 1=mheg off
				    return 0;
				case 6:
                    exTAP_MHEG_Disable(0);      //0=mheg on, 1=mheg off
				    return 0;
				case 7:
				    close();
				    TAP_Exit();
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            return param1; // cascades key to other TAP's
        case RKEY_Exit:
			DeleteMenu(menu);
            return 0; // exit without Cascading
        default:
	        TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
    		return 0;
    }
    return param1;
}

int key_handler(dword param1)
{
    switch(param1) {
        case RKEY_Power:
            if (in_shutdown == 0) {
                //RKEY_Power event manually generated
            }
            else
            if (in_shutdown == 1) {
                TAP_GenerateEvent(EVT_KEY, RKEY_Power, 0);
                in_shutdown ++;
            }
            else
            if (in_shutdown > 1) {
//                exTAP_CallFirmware((void *)FW_SHUTDOWN_IRQ_ENABLE_1209, 0, 0,0,0);    // renable shutdown irq - not necessary since we will force a shutdown
                exTAP_Shutdown();
            }
            break;

        case RKEY_Recall:                 //menu
            ShowMenu(&menu);
            return 0;
    }
    return param1;
}

//----------------------------------------------------------------------------
int init(void)
{
    int ret;
    menu.visible = 0;
    load_tap_menu.visible = 0;

    ret = apply_hooks();
    if (ret == 0) strcpy(menu_title, "Shutdown is Enabled");

    return 0;
}

int close(void)
{
    TAP_Print("\nclosing %s",__tap_program_name__);
    ShowMessage("closing", (char *)__tap_program_name__, 100, 1);

    //all hooks MUST be removed before shutting down a tap
    remove_hooks();
    //patches may be left enabled
    exTAP_MHEG_Disable(0);      //undo MHEG patch

    DeleteMenu(&load_tap_menu);
    DeleteMenu(&menu);
    return 0;
}

//----------------------------------------------------------------------------
//generic message functions

dword ShowMessage(char* lpMessage, char* lpMessage1, dword dwDelay, int wait)
{
	dword rgn;															// stores rgn-handle for osd
	dword w;															// stores width of message-text

	rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE );					// create rgn-handle
	w = TAP_Osd_GetW(lpMessage, 0, FNT_Size_1622 ) + 10;				// calculate width of message
	if (TAP_Osd_GetW(lpMessage1, 0, FNT_Size_1622 ) + 10>w) {
		w = TAP_Osd_GetW(lpMessage1, 0, FNT_Size_1622 ) + 10;			// if second message is larger, calculate new width of message
    }

	if (w > 720) w = 720;												// if message is too long
	TAP_Osd_FillBox(rgn, (720-w)/2-5, 265, w+10, 54, RGB(19,19,19) );	// draw background-box for border

	TAP_Osd_PutS(rgn, (720-w)/2, 270, (720+w)/2, lpMessage,
		COLOR_White, RGB(3,5,10), 0, FNT_Size_1622, FALSE, ALIGN_CENTER);	 // show 1. message
	TAP_Osd_PutS(rgn, (720-w)/2, 270+22, (720+w)/2, lpMessage1,
		COLOR_White, RGB(3,5,10), 0, FNT_Size_1622, FALSE, ALIGN_CENTER);	 // show 2. message
	if (wait) {
        message_showing = 0;
	    TAP_Delay(dwDelay);												// show it for dwDelay /100 seconds
	    TAP_Osd_Delete(rgn);											// release rgn-handle
        message_rgn = 0;
	}
	else {
        message_showing = TAP_GetTick() + dwDelay;
        message_rgn = rgn;
    	return rgn;
	}
	return 0;
}

void ShowMenu(type_menu *menu)
{
    int i;

	if (!menu->visible) {
	    menu->rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);
	    TAP_Win_SetDefaultColor(&menu->win);
	    TAP_Win_Create(&menu->win, menu->rgn, 250, 210, 200, 80, FALSE, FALSE);
        TAP_Win_SetTitle(&menu->win, ptrfix(menu->fixme_titlep), 0, FNT_Size_1622);
	    for (i=1; i<=menu->item[0].id; i++) {
    	    TAP_Win_AddItem(&menu->win, menu->item[i].str);
	    }
	    TAP_Win_SetSelection(&menu->win, 0);
    menu->visible = 1;
	}
}

void DeleteMenu(type_menu *menu)
{
	if (menu->visible) {
	    TAP_Win_Delete(&menu->win);
    	TAP_Osd_Delete(menu->rgn);
	    menu->visible = 0;
    }
}

//----------------------------------------------------------------------------
//load tap funtions

const char load_tap_title[] = "Load TAP";

type_menu load_tap_menu = {
    0,
    0,
    {0},
    0,
    (char *)load_tap_title,
    {
        {1, "xxx",                  },
        {0, "Back",                 },
    },
};

int load_tap_menu_fill()
{
    TYPE_File file;
    int num_files;
    int i,j;

    //assume we are in the ProgramFiles directory
    num_files = TAP_Hdd_FindFirst(&file);
    TAP_Print("\nfound %d files", num_files);

    for (i=0,j=1; (i<num_files) && (j<0x20); i++) {
        if (strstr(file.name, ".tap")) {
            TAP_Print("\nfound file: %s", file.name);
            load_tap_menu.item[j].id = j-1;
            strncpy(load_tap_menu.item[j].str, file.name, 0x20);
            j++;
        }

        TAP_Hdd_FindNext(&file);
    }
    TAP_Print("\nfound %d taps", j);

    load_tap_menu.item[j].id = j-1;
    strncpy(load_tap_menu.item[j].str, "Cancel", 0x20);
    j++;

    load_tap_menu.item[0].id = j-1;

    return 0;
}

int load_tap_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int ret;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);

            //get the tap filename
            if (selection < (load_tap_menu.item[0].id-1)) {
                //load the tap
                TAP_Print("\nselection: %d", selection);
                TAP_Print("\nLoading TAP: %s", load_tap_menu.item[selection+1].str);
                ret = exTAP_TAP_Load(load_tap_menu.item[selection+1].str);
                if (!ret) TAP_Print("\nexTAP_TAP_Load() failed");
                else TAP_Print("\nexTAP_TAP_Load() ok, ret=%x", ret);
            }
            return 0;
            break;
        case RKEY_Recall:
            DeleteMenu(menu);
            return param1; // cascades key to other TAP's
        case exRKEY_Exit:
            DeleteMenu(menu);
            return 0; // exit without Cascading
        default:
            TAP_Win_Action(&menu->win, param1); // send all other key's to menu-Win
            return 0;
    }
    return param1;
}


//-----------------------------------------------------------------------------
//shutdown functions

//shutdown_task
//80108C0C 27 BD FF B8   addiu   $sp, -0x48
//80108C10 AF B4 00 38   sw      $s4, 0x38($sp)

type_hook hook_patch[] = {
//   enable,address,    length, original_data
    {1,     0x80108c0c, 8,      {0x27, 0xbd, 0xff, 0xb8,  0xaf, 0xb4, 0x00, 0x38, }, },  //shutdown
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //

    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
    {0,     0x00000000, 8,      {0x00, 0x00, 0x00, 0x00,  0x00, 0x00, 0x00, 0x00, }, },  //
};

int apply_hooks(void)
{
    int i;
    int ret;

    for (i=0; i<MAX_HOOKS; i++) {
        if (hook_patch[i].enable) {
            TAP_Print("\nApplying Hook %d", i);
            hook_id[i] = exTAP_ApplyHook(&hook_patch[i]);
        }
        else {
            hook_id[i] = -1;
        }
    }
    return 0;
}

int remove_hooks(void)
{
    int i;
    int ret;

    for (i=0; i<MAX_HOOKS; i++) {
        if (hook_id[i] >=0) {
            TAP_Print("\nRemoving Hook %d", i);
            ret = exTAP_RemoveHook(hook_id[i]);
            if (ret == 0) hook_id[i] = -1;
        }
    }
    return 0;
}

dword shutdown_hook(word event, dword param1, dword param2)
{
//    type_hook_registers *reg;
//
//    reg = (type_hook_registers *)param2;
//    TAP_Print("\n 0=%08x  t0=%08x  s0=%08x  t8=%08x", reg->zero, reg->t0, reg->s0, reg->t8);
//    TAP_Print("\nat=%08x  t1=%08x  s1=%08x  t9=%08x", reg->at,   reg->t1, reg->s1, reg->t9);
//    TAP_Print("\nv0=%08x  t2=%08x  s2=%08x  k0=%08x", reg->v0,   reg->t2, reg->s2, reg->k0);
//    TAP_Print("\nv1=%08x  t3=%08x  s3=%08x  k1=%08x", reg->v1,   reg->t3, reg->s3, reg->k1);
//    TAP_Print("\na0=%08x  t4=%08x  s4=%08x  gp=%08x", reg->a0,   reg->t4, reg->s4, reg->gp);
//    TAP_Print("\na1=%08x  t5=%08x  s5=%08x  sp=%08x", reg->a1,   reg->t5, reg->s5, reg->sp);
//    TAP_Print("\na2=%08x  t6=%08x  s6=%08x  s8=%08x", reg->a2,   reg->t6, reg->s6, reg->s8);
//    TAP_Print("\na3=%08x  t7=%08x  s7=%08x  ra=%08x", reg->a3,   reg->t7, reg->s7, reg->ra);
//    TAP_Print("\n");
//    if ((reg->a0 > 0x80000000) && (reg->a0 < 0x84000000)) {
//        TAP_Print("\na0 = %08x", reg->a0);
//        dump(reg->a0, 0x40);
//    }
//    if ((reg->a1 > 0x80000000) && (reg->a1 < 0x84000000)) {
//        TAP_Print("\na1 = %08x", reg->a1);
//        dump(reg->a1, 0x40);
//    }
//    if ((reg->a2 > 0x80000000) && (reg->a2 < 0x84000000)) {
//        TAP_Print("\na2 = %08x", reg->a2);
//        dump(reg->a2, 0x40);
//    }
//    if ((reg->a3 > 0x80000000) && (reg->a3 < 0x84000000)) {
//        TAP_Print("\na3 = %08x", reg->a3);
//        dump(reg->a3, 0x40);
//    }

    if (in_shutdown == 0) {
        in_shutdown ++;
        exTAP_CallFirmware((void *)FW_SHUTDOWN_IRQ_DISABLE_1209, 0, 0,0,0);       //disable shutdown irq, prevents automatic shutdown on timeout
        TAP_GenerateEvent(EVT_KEY, RKEY_Power, 0);
        return -1;                  //force rts; will not shutdown
    }
    //need this to ensure the RKEY_Power completes ...
    else
    if (in_shutdown == 1) {
        in_shutdown ++;
        return -1;                  //force rts; will not shutdown
    }
    else {
        in_shutdown = 0;
        return param1;              //allow continue; will shutdown
    }
    return param1;
}

//----------------------------------------------------------------------------


